{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch \n",
    "import torch.nn as nn\n",
    "import torchvision\n",
    "import torchvision.transforms as transforms"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Device configuration\n",
    "device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')\n",
    "\n",
    "# Hyper-parameters\n",
    "sequence_length = 28\n",
    "input_size = 28\n",
    "hidden_size = 128\n",
    "num_layers = 2\n",
    "num_classes = 10\n",
    "batch_size = 100\n",
    "num_epochs = 50\n",
    "learning_rate = 0.003"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "# MNIST dataset\n",
    "train_dataset = torchvision.datasets.MNIST(root='../../data/',\n",
    "                                           train=True, \n",
    "                                           transform=transforms.ToTensor(),\n",
    "                                           download=True)\n",
    "\n",
    "test_dataset = torchvision.datasets.MNIST(root='../../data/',\n",
    "                                          train=False, \n",
    "                                          transform=transforms.ToTensor())\n",
    "\n",
    "# Data loader\n",
    "train_loader = torch.utils.data.DataLoader(dataset=train_dataset,\n",
    "                                           batch_size=batch_size, \n",
    "                                           shuffle=True)\n",
    "\n",
    "test_loader = torch.utils.data.DataLoader(dataset=test_dataset,\n",
    "                                          batch_size=batch_size, \n",
    "                                          shuffle=False)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Bidirectional recurrent neural network (many-to-one)\n",
    "class BiRNN(nn.Module):\n",
    "    def __init__(self, input_size, hidden_size, num_layers, num_classes):\n",
    "        super(BiRNN, self).__init__()\n",
    "        self.hidden_size = hidden_size\n",
    "        self.num_layers = num_layers\n",
    "        self.lstm = nn.LSTM(input_size, hidden_size, num_layers, batch_first=True, bidirectional=True)\n",
    "        self.fc = nn.Linear(hidden_size*2, num_classes)  # 2 for bidirection\n",
    "    \n",
    "    def forward(self, x):\n",
    "        # Set initial states\n",
    "        h0 = torch.zeros(self.num_layers*2, x.size(0), self.hidden_size).to(device) # 2 for bidirection \n",
    "        c0 = torch.zeros(self.num_layers*2, x.size(0), self.hidden_size).to(device)\n",
    "        \n",
    "        # Forward propagate LSTM\n",
    "        out, _ = self.lstm(x, (h0, c0))  # out: tensor of shape (batch_size, seq_length, hidden_size*2)\n",
    "        \n",
    "        # Decode the hidden state of the last time step\n",
    "        out = self.fc(out[:, -1, :])\n",
    "        return out"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "model = BiRNN(input_size, hidden_size, num_layers, num_classes).to(device)\n",
    "\n",
    "# Loss and optimizer\n",
    "criterion = nn.CrossEntropyLoss()\n",
    "optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Epoch [1/50], Step [100/600], Loss: 0.0669\n",
      "Epoch [1/50], Step [200/600], Loss: 0.1369\n",
      "Epoch [1/50], Step [300/600], Loss: 0.1123\n",
      "Epoch [1/50], Step [400/600], Loss: 0.0328\n",
      "Epoch [1/50], Step [500/600], Loss: 0.0863\n",
      "Epoch [1/50], Step [600/600], Loss: 0.0191\n",
      "Epoch [2/50], Step [100/600], Loss: 0.0391\n",
      "Epoch [2/50], Step [200/600], Loss: 0.0276\n",
      "Epoch [2/50], Step [300/600], Loss: 0.1112\n",
      "Epoch [2/50], Step [400/600], Loss: 0.0299\n",
      "Epoch [2/50], Step [500/600], Loss: 0.0371\n",
      "Epoch [2/50], Step [600/600], Loss: 0.0344\n",
      "Epoch [3/50], Step [100/600], Loss: 0.0225\n",
      "Epoch [3/50], Step [200/600], Loss: 0.0926\n",
      "Epoch [3/50], Step [300/600], Loss: 0.0395\n",
      "Epoch [3/50], Step [400/600], Loss: 0.0414\n",
      "Epoch [3/50], Step [500/600], Loss: 0.0250\n",
      "Epoch [3/50], Step [600/600], Loss: 0.0391\n",
      "Epoch [4/50], Step [100/600], Loss: 0.0261\n",
      "Epoch [4/50], Step [200/600], Loss: 0.1638\n",
      "Epoch [4/50], Step [300/600], Loss: 0.0159\n",
      "Epoch [4/50], Step [400/600], Loss: 0.0343\n",
      "Epoch [4/50], Step [500/600], Loss: 0.0125\n",
      "Epoch [4/50], Step [600/600], Loss: 0.0615\n",
      "Epoch [5/50], Step [100/600], Loss: 0.0110\n",
      "Epoch [5/50], Step [200/600], Loss: 0.1151\n",
      "Epoch [5/50], Step [300/600], Loss: 0.0260\n",
      "Epoch [5/50], Step [400/600], Loss: 0.0410\n",
      "Epoch [5/50], Step [500/600], Loss: 0.0508\n",
      "Epoch [5/50], Step [600/600], Loss: 0.0447\n",
      "Epoch [6/50], Step [100/600], Loss: 0.0194\n",
      "Epoch [6/50], Step [200/600], Loss: 0.0195\n",
      "Epoch [6/50], Step [300/600], Loss: 0.0422\n",
      "Epoch [6/50], Step [400/600], Loss: 0.0015\n",
      "Epoch [6/50], Step [500/600], Loss: 0.0381\n",
      "Epoch [6/50], Step [600/600], Loss: 0.0428\n",
      "Epoch [7/50], Step [100/600], Loss: 0.0352\n",
      "Epoch [7/50], Step [200/600], Loss: 0.0141\n",
      "Epoch [7/50], Step [300/600], Loss: 0.0243\n",
      "Epoch [7/50], Step [400/600], Loss: 0.0202\n",
      "Epoch [7/50], Step [500/600], Loss: 0.0446\n",
      "Epoch [7/50], Step [600/600], Loss: 0.0025\n",
      "Epoch [8/50], Step [100/600], Loss: 0.0007\n",
      "Epoch [8/50], Step [200/600], Loss: 0.0508\n",
      "Epoch [8/50], Step [300/600], Loss: 0.0598\n",
      "Epoch [8/50], Step [400/600], Loss: 0.0300\n",
      "Epoch [8/50], Step [500/600], Loss: 0.1283\n",
      "Epoch [8/50], Step [600/600], Loss: 0.0011\n",
      "Epoch [9/50], Step [100/600], Loss: 0.0314\n",
      "Epoch [9/50], Step [200/600], Loss: 0.0910\n",
      "Epoch [9/50], Step [300/600], Loss: 0.0639\n",
      "Epoch [9/50], Step [400/600], Loss: 0.0135\n",
      "Epoch [9/50], Step [500/600], Loss: 0.0362\n",
      "Epoch [9/50], Step [600/600], Loss: 0.0012\n",
      "Epoch [10/50], Step [100/600], Loss: 0.0454\n",
      "Epoch [10/50], Step [200/600], Loss: 0.0019\n",
      "Epoch [10/50], Step [300/600], Loss: 0.0077\n",
      "Epoch [10/50], Step [400/600], Loss: 0.0627\n",
      "Epoch [10/50], Step [500/600], Loss: 0.0291\n",
      "Epoch [10/50], Step [600/600], Loss: 0.0062\n",
      "Epoch [11/50], Step [100/600], Loss: 0.0188\n",
      "Epoch [11/50], Step [200/600], Loss: 0.0018\n",
      "Epoch [11/50], Step [300/600], Loss: 0.0048\n",
      "Epoch [11/50], Step [400/600], Loss: 0.0434\n",
      "Epoch [11/50], Step [500/600], Loss: 0.0169\n",
      "Epoch [11/50], Step [600/600], Loss: 0.0391\n",
      "Epoch [12/50], Step [100/600], Loss: 0.0022\n",
      "Epoch [12/50], Step [200/600], Loss: 0.0029\n",
      "Epoch [12/50], Step [300/600], Loss: 0.0268\n",
      "Epoch [12/50], Step [400/600], Loss: 0.0362\n",
      "Epoch [12/50], Step [500/600], Loss: 0.0021\n",
      "Epoch [12/50], Step [600/600], Loss: 0.0214\n",
      "Epoch [13/50], Step [100/600], Loss: 0.0231\n",
      "Epoch [13/50], Step [200/600], Loss: 0.0212\n",
      "Epoch [13/50], Step [300/600], Loss: 0.0013\n",
      "Epoch [13/50], Step [400/600], Loss: 0.0131\n",
      "Epoch [13/50], Step [500/600], Loss: 0.0099\n",
      "Epoch [13/50], Step [600/600], Loss: 0.0147\n",
      "Epoch [14/50], Step [100/600], Loss: 0.0612\n",
      "Epoch [14/50], Step [200/600], Loss: 0.0015\n",
      "Epoch [14/50], Step [300/600], Loss: 0.0118\n",
      "Epoch [14/50], Step [400/600], Loss: 0.0013\n",
      "Epoch [14/50], Step [500/600], Loss: 0.0116\n",
      "Epoch [14/50], Step [600/600], Loss: 0.0489\n",
      "Epoch [15/50], Step [100/600], Loss: 0.0004\n",
      "Epoch [15/50], Step [200/600], Loss: 0.0199\n",
      "Epoch [15/50], Step [300/600], Loss: 0.0013\n",
      "Epoch [15/50], Step [400/600], Loss: 0.0068\n",
      "Epoch [15/50], Step [500/600], Loss: 0.0352\n",
      "Epoch [15/50], Step [600/600], Loss: 0.0080\n",
      "Epoch [16/50], Step [100/600], Loss: 0.0128\n",
      "Epoch [16/50], Step [200/600], Loss: 0.0120\n",
      "Epoch [16/50], Step [300/600], Loss: 0.0686\n",
      "Epoch [16/50], Step [400/600], Loss: 0.0275\n",
      "Epoch [16/50], Step [500/600], Loss: 0.0152\n",
      "Epoch [16/50], Step [600/600], Loss: 0.0013\n",
      "Epoch [17/50], Step [100/600], Loss: 0.0029\n",
      "Epoch [17/50], Step [200/600], Loss: 0.0017\n",
      "Epoch [17/50], Step [300/600], Loss: 0.0005\n",
      "Epoch [17/50], Step [400/600], Loss: 0.0004\n",
      "Epoch [17/50], Step [500/600], Loss: 0.0045\n",
      "Epoch [17/50], Step [600/600], Loss: 0.0082\n",
      "Epoch [18/50], Step [100/600], Loss: 0.0038\n",
      "Epoch [18/50], Step [200/600], Loss: 0.0037\n",
      "Epoch [18/50], Step [300/600], Loss: 0.0036\n",
      "Epoch [18/50], Step [400/600], Loss: 0.0343\n",
      "Epoch [18/50], Step [500/600], Loss: 0.0074\n",
      "Epoch [18/50], Step [600/600], Loss: 0.0076\n",
      "Epoch [19/50], Step [100/600], Loss: 0.0002\n",
      "Epoch [19/50], Step [200/600], Loss: 0.0519\n",
      "Epoch [19/50], Step [300/600], Loss: 0.0033\n",
      "Epoch [19/50], Step [400/600], Loss: 0.0005\n",
      "Epoch [19/50], Step [500/600], Loss: 0.0169\n",
      "Epoch [19/50], Step [600/600], Loss: 0.0177\n",
      "Epoch [20/50], Step [100/600], Loss: 0.0075\n",
      "Epoch [20/50], Step [200/600], Loss: 0.0010\n",
      "Epoch [20/50], Step [300/600], Loss: 0.0068\n",
      "Epoch [20/50], Step [400/600], Loss: 0.0201\n",
      "Epoch [20/50], Step [500/600], Loss: 0.0018\n",
      "Epoch [20/50], Step [600/600], Loss: 0.0267\n",
      "Epoch [21/50], Step [100/600], Loss: 0.0009\n",
      "Epoch [21/50], Step [200/600], Loss: 0.0008\n",
      "Epoch [21/50], Step [300/600], Loss: 0.0054\n",
      "Epoch [21/50], Step [400/600], Loss: 0.0013\n",
      "Epoch [21/50], Step [500/600], Loss: 0.0347\n",
      "Epoch [21/50], Step [600/600], Loss: 0.0051\n",
      "Epoch [22/50], Step [100/600], Loss: 0.0053\n",
      "Epoch [22/50], Step [200/600], Loss: 0.0236\n",
      "Epoch [22/50], Step [300/600], Loss: 0.0002\n",
      "Epoch [22/50], Step [400/600], Loss: 0.0038\n",
      "Epoch [22/50], Step [500/600], Loss: 0.0329\n",
      "Epoch [22/50], Step [600/600], Loss: 0.0121\n",
      "Epoch [23/50], Step [100/600], Loss: 0.0001\n",
      "Epoch [23/50], Step [200/600], Loss: 0.0053\n",
      "Epoch [23/50], Step [300/600], Loss: 0.0007\n",
      "Epoch [23/50], Step [400/600], Loss: 0.0003\n",
      "Epoch [23/50], Step [500/600], Loss: 0.0053\n",
      "Epoch [23/50], Step [600/600], Loss: 0.0001\n",
      "Epoch [24/50], Step [100/600], Loss: 0.0004\n",
      "Epoch [24/50], Step [200/600], Loss: 0.0395\n",
      "Epoch [24/50], Step [300/600], Loss: 0.0156\n",
      "Epoch [24/50], Step [400/600], Loss: 0.0423\n",
      "Epoch [24/50], Step [500/600], Loss: 0.0089\n",
      "Epoch [24/50], Step [600/600], Loss: 0.0002\n",
      "Epoch [25/50], Step [100/600], Loss: 0.0025\n",
      "Epoch [25/50], Step [200/600], Loss: 0.0187\n",
      "Epoch [25/50], Step [300/600], Loss: 0.0004\n",
      "Epoch [25/50], Step [400/600], Loss: 0.0009\n",
      "Epoch [25/50], Step [500/600], Loss: 0.0001\n",
      "Epoch [25/50], Step [600/600], Loss: 0.0037\n",
      "Epoch [26/50], Step [100/600], Loss: 0.0004\n",
      "Epoch [26/50], Step [200/600], Loss: 0.0009\n",
      "Epoch [26/50], Step [300/600], Loss: 0.0011\n",
      "Epoch [26/50], Step [400/600], Loss: 0.0009\n",
      "Epoch [26/50], Step [500/600], Loss: 0.0016\n",
      "Epoch [26/50], Step [600/600], Loss: 0.0425\n",
      "Epoch [27/50], Step [100/600], Loss: 0.0003\n",
      "Epoch [27/50], Step [200/600], Loss: 0.0012\n",
      "Epoch [27/50], Step [300/600], Loss: 0.0005\n",
      "Epoch [27/50], Step [400/600], Loss: 0.0198\n",
      "Epoch [27/50], Step [500/600], Loss: 0.0063\n",
      "Epoch [27/50], Step [600/600], Loss: 0.0316\n",
      "Epoch [28/50], Step [100/600], Loss: 0.0008\n",
      "Epoch [28/50], Step [200/600], Loss: 0.0002\n",
      "Epoch [28/50], Step [300/600], Loss: 0.0007\n",
      "Epoch [28/50], Step [400/600], Loss: 0.0009\n",
      "Epoch [28/50], Step [500/600], Loss: 0.0010\n",
      "Epoch [28/50], Step [600/600], Loss: 0.0003\n",
      "Epoch [29/50], Step [100/600], Loss: 0.0014\n",
      "Epoch [29/50], Step [200/600], Loss: 0.0052\n",
      "Epoch [29/50], Step [300/600], Loss: 0.0367\n",
      "Epoch [29/50], Step [400/600], Loss: 0.0003\n",
      "Epoch [29/50], Step [500/600], Loss: 0.0003\n",
      "Epoch [29/50], Step [600/600], Loss: 0.0007\n",
      "Epoch [30/50], Step [100/600], Loss: 0.0009\n",
      "Epoch [30/50], Step [200/600], Loss: 0.0003\n",
      "Epoch [30/50], Step [300/600], Loss: 0.0001\n",
      "Epoch [30/50], Step [400/600], Loss: 0.0143\n",
      "Epoch [30/50], Step [500/600], Loss: 0.0047\n",
      "Epoch [30/50], Step [600/600], Loss: 0.0048\n",
      "Epoch [31/50], Step [100/600], Loss: 0.0000\n",
      "Epoch [31/50], Step [200/600], Loss: 0.0004\n",
      "Epoch [31/50], Step [300/600], Loss: 0.0048\n",
      "Epoch [31/50], Step [400/600], Loss: 0.0003\n",
      "Epoch [31/50], Step [500/600], Loss: 0.0007\n",
      "Epoch [31/50], Step [600/600], Loss: 0.0011\n",
      "Epoch [32/50], Step [100/600], Loss: 0.0035\n",
      "Epoch [32/50], Step [200/600], Loss: 0.0018\n",
      "Epoch [32/50], Step [300/600], Loss: 0.0010\n",
      "Epoch [32/50], Step [400/600], Loss: 0.0013\n",
      "Epoch [32/50], Step [500/600], Loss: 0.0109\n",
      "Epoch [32/50], Step [600/600], Loss: 0.0685\n",
      "Epoch [33/50], Step [100/600], Loss: 0.0003\n",
      "Epoch [33/50], Step [200/600], Loss: 0.0006\n",
      "Epoch [33/50], Step [300/600], Loss: 0.0002\n",
      "Epoch [33/50], Step [400/600], Loss: 0.0000\n",
      "Epoch [33/50], Step [500/600], Loss: 0.0048\n",
      "Epoch [33/50], Step [600/600], Loss: 0.0002\n",
      "Epoch [34/50], Step [100/600], Loss: 0.0110\n",
      "Epoch [34/50], Step [200/600], Loss: 0.0012\n",
      "Epoch [34/50], Step [300/600], Loss: 0.0024\n",
      "Epoch [34/50], Step [400/600], Loss: 0.0002\n",
      "Epoch [34/50], Step [500/600], Loss: 0.0245\n",
      "Epoch [34/50], Step [600/600], Loss: 0.0001\n",
      "Epoch [35/50], Step [100/600], Loss: 0.0002\n",
      "Epoch [35/50], Step [200/600], Loss: 0.0299\n",
      "Epoch [35/50], Step [300/600], Loss: 0.0416\n",
      "Epoch [35/50], Step [400/600], Loss: 0.0007\n",
      "Epoch [35/50], Step [500/600], Loss: 0.0003\n",
      "Epoch [35/50], Step [600/600], Loss: 0.0005\n",
      "Epoch [36/50], Step [100/600], Loss: 0.0001\n",
      "Epoch [36/50], Step [200/600], Loss: 0.0006\n",
      "Epoch [36/50], Step [300/600], Loss: 0.0000\n",
      "Epoch [36/50], Step [400/600], Loss: 0.0000\n",
      "Epoch [36/50], Step [500/600], Loss: 0.0011\n",
      "Epoch [36/50], Step [600/600], Loss: 0.0001\n",
      "Epoch [37/50], Step [100/600], Loss: 0.0000\n",
      "Epoch [37/50], Step [200/600], Loss: 0.0011\n",
      "Epoch [37/50], Step [300/600], Loss: 0.0004\n",
      "Epoch [37/50], Step [400/600], Loss: 0.0005\n",
      "Epoch [37/50], Step [500/600], Loss: 0.0004\n",
      "Epoch [37/50], Step [600/600], Loss: 0.0024\n",
      "Epoch [38/50], Step [100/600], Loss: 0.0006\n",
      "Epoch [38/50], Step [200/600], Loss: 0.0003\n",
      "Epoch [38/50], Step [300/600], Loss: 0.0037\n",
      "Epoch [38/50], Step [400/600], Loss: 0.0551\n",
      "Epoch [38/50], Step [500/600], Loss: 0.0009\n",
      "Epoch [38/50], Step [600/600], Loss: 0.1011\n",
      "Epoch [39/50], Step [100/600], Loss: 0.0031\n",
      "Epoch [39/50], Step [200/600], Loss: 0.0005\n",
      "Epoch [39/50], Step [300/600], Loss: 0.0001\n",
      "Epoch [39/50], Step [400/600], Loss: 0.1076\n",
      "Epoch [39/50], Step [500/600], Loss: 0.0002\n",
      "Epoch [39/50], Step [600/600], Loss: 0.0003\n",
      "Epoch [40/50], Step [100/600], Loss: 0.0001\n",
      "Epoch [40/50], Step [200/600], Loss: 0.0004\n",
      "Epoch [40/50], Step [300/600], Loss: 0.0001\n",
      "Epoch [40/50], Step [400/600], Loss: 0.0000\n",
      "Epoch [40/50], Step [500/600], Loss: 0.0001\n",
      "Epoch [40/50], Step [600/600], Loss: 0.0041\n",
      "Epoch [41/50], Step [100/600], Loss: 0.0029\n",
      "Epoch [41/50], Step [200/600], Loss: 0.0007\n",
      "Epoch [41/50], Step [300/600], Loss: 0.0164\n",
      "Epoch [41/50], Step [400/600], Loss: 0.0010\n",
      "Epoch [41/50], Step [500/600], Loss: 0.0018\n",
      "Epoch [41/50], Step [600/600], Loss: 0.0000\n",
      "Epoch [42/50], Step [100/600], Loss: 0.0113\n",
      "Epoch [42/50], Step [200/600], Loss: 0.0048\n",
      "Epoch [42/50], Step [300/600], Loss: 0.0370\n",
      "Epoch [42/50], Step [400/600], Loss: 0.0399\n",
      "Epoch [42/50], Step [500/600], Loss: 0.0199\n",
      "Epoch [42/50], Step [600/600], Loss: 0.0171\n",
      "Epoch [43/50], Step [100/600], Loss: 0.0608\n",
      "Epoch [43/50], Step [200/600], Loss: 0.0004\n",
      "Epoch [43/50], Step [300/600], Loss: 0.0001\n",
      "Epoch [43/50], Step [400/600], Loss: 0.0008\n",
      "Epoch [43/50], Step [500/600], Loss: 0.0002\n",
      "Epoch [43/50], Step [600/600], Loss: 0.0015\n",
      "Epoch [44/50], Step [100/600], Loss: 0.0057\n",
      "Epoch [44/50], Step [200/600], Loss: 0.0007\n",
      "Epoch [44/50], Step [300/600], Loss: 0.0012\n",
      "Epoch [44/50], Step [400/600], Loss: 0.0054\n",
      "Epoch [44/50], Step [500/600], Loss: 0.0111\n",
      "Epoch [44/50], Step [600/600], Loss: 0.0001\n",
      "Epoch [45/50], Step [100/600], Loss: 0.0012\n",
      "Epoch [45/50], Step [200/600], Loss: 0.0017\n",
      "Epoch [45/50], Step [300/600], Loss: 0.0003\n",
      "Epoch [45/50], Step [400/600], Loss: 0.0401\n",
      "Epoch [45/50], Step [500/600], Loss: 0.0005\n",
      "Epoch [45/50], Step [600/600], Loss: 0.0635\n",
      "Epoch [46/50], Step [100/600], Loss: 0.0008\n",
      "Epoch [46/50], Step [200/600], Loss: 0.0167\n",
      "Epoch [46/50], Step [300/600], Loss: 0.0422\n",
      "Epoch [46/50], Step [400/600], Loss: 0.0127\n",
      "Epoch [46/50], Step [500/600], Loss: 0.0001\n",
      "Epoch [46/50], Step [600/600], Loss: 0.0006\n",
      "Epoch [47/50], Step [100/600], Loss: 0.0000\n",
      "Epoch [47/50], Step [200/600], Loss: 0.0139\n",
      "Epoch [47/50], Step [300/600], Loss: 0.0005\n",
      "Epoch [47/50], Step [400/600], Loss: 0.0284\n",
      "Epoch [47/50], Step [500/600], Loss: 0.0012\n",
      "Epoch [47/50], Step [600/600], Loss: 0.0361\n",
      "Epoch [48/50], Step [100/600], Loss: 0.0001\n",
      "Epoch [48/50], Step [200/600], Loss: 0.0002\n",
      "Epoch [48/50], Step [300/600], Loss: 0.0006\n",
      "Epoch [48/50], Step [400/600], Loss: 0.0037\n",
      "Epoch [48/50], Step [500/600], Loss: 0.0003\n",
      "Epoch [48/50], Step [600/600], Loss: 0.0001\n",
      "Epoch [49/50], Step [100/600], Loss: 0.0004\n",
      "Epoch [49/50], Step [200/600], Loss: 0.0005\n",
      "Epoch [49/50], Step [300/600], Loss: 0.0001\n",
      "Epoch [49/50], Step [400/600], Loss: 0.0064\n",
      "Epoch [49/50], Step [500/600], Loss: 0.0001\n",
      "Epoch [49/50], Step [600/600], Loss: 0.0002\n",
      "Epoch [50/50], Step [100/600], Loss: 0.0001\n",
      "Epoch [50/50], Step [200/600], Loss: 0.0055\n",
      "Epoch [50/50], Step [300/600], Loss: 0.0015\n",
      "Epoch [50/50], Step [400/600], Loss: 0.0002\n",
      "Epoch [50/50], Step [500/600], Loss: 0.0003\n",
      "Epoch [50/50], Step [600/600], Loss: 0.0048\n"
     ]
    }
   ],
   "source": [
    "# Train the model\n",
    "total_step = len(train_loader)\n",
    "for epoch in range(num_epochs):\n",
    "    for i, (images, labels) in enumerate(train_loader):\n",
    "        images = images.reshape(-1, sequence_length, input_size).to(device)\n",
    "        labels = labels.to(device)\n",
    "        \n",
    "        # Forward pass\n",
    "        outputs = model(images)\n",
    "        loss = criterion(outputs, labels)\n",
    "        \n",
    "        # Backward and optimize\n",
    "        optimizer.zero_grad()\n",
    "        loss.backward()\n",
    "        optimizer.step()\n",
    "        \n",
    "        if (i+1) % 100 == 0:\n",
    "            print ('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}' \n",
    "                   .format(epoch+1, num_epochs, i+1, total_step, loss.item()))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Test Accuracy of the model on the 10000 test images: 99.03 %\n"
     ]
    }
   ],
   "source": [
    "# Test the model\n",
    "with torch.no_grad():\n",
    "    correct = 0\n",
    "    total = 0\n",
    "    for images, labels in test_loader:\n",
    "        images = images.reshape(-1, sequence_length, input_size).to(device)\n",
    "        labels = labels.to(device)\n",
    "        outputs = model(images)\n",
    "        _, predicted = torch.max(outputs.data, 1)\n",
    "        total += labels.size(0)\n",
    "        correct += (predicted == labels).sum().item()\n",
    "\n",
    "    print('Test Accuracy of the model on the 10000 test images: {} %'.format(100 * correct / total)) \n",
    "\n",
    "# Save the model checkpoint\n",
    "torch.save(model.state_dict(), 'bi_lstm.ckpt')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "772508630a427619dd49b8cfe9c2ae7d9dbaeae5c7cd5fb2ab9db30424e7636a"
  },
  "kernelspec": {
   "display_name": "Python 3.6.13 64-bit ('python36': conda)",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.13"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
